#ifdef CP_CHANGES
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef DEBUG
#include <assert.h>
#endif


#include <stdlib.h>
#include <ctype.h>

#include "rules.h"
#include "decode.h"
#include "debug.h"
#include "plugbase.h"
#include "parser.h"
#include "plugin_enum.h"
#include "util.h"
#include "mstring.h"
#include "event.h"

#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <Policyd.H>
#endif /* !WIN32 */

#ifdef UTM
#include "inline.h"  //For direction marcos
#endif /* UTM */

#ifdef STM
#define POLICYD_TIMEOUT    1
#endif /* STM */

typedef struct _QueryPolicyData
{
    char policy_protocol[20];
    int default_mode;

} QueryPolicyData;

void *instance = NULL;

void QueryPolicyInit(char *, OptTreeNode *, int);
void ParseQuery(char *, OptTreeNode *);
int QueryPolicy(Packet *, struct _OptTreeNode *, OptFpList *);

void SetupQueryPolicy(void)
{
    RegisterPlugin("query", QueryPolicyInit);
    
    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Plugin: QueryPolicy Initialized\n"););
}

void QueryPolicyInit(char *data, OptTreeNode *otn, int protocol)
{
    Event *event;
    if(otn->ds_list[PLUGIN_QUERYPOLICY])
    {
        FatalError("%s(%d): Multiple query options in rule\n", file_name,
                file_line);
    }

    otn->ds_list[PLUGIN_QUERYPOLICY] = (QueryPolicyData *)
          SnortAlloc(sizeof(QueryPolicyData));

    ParseQuery(data, otn);
    if(instance == NULL)
    {
        POLICYD_RET_CODE ret = createPolicyClientInstance(&instance);
        if(ret != POLICYD_RET_OK)
        {
            instance = NULL;
            ErrorMessage("%s(%d): Create policyd client instance failed, %d\n", 
                          __FILE__, __LINE__, ret);
        }
#ifdef STM
        else
        {
            /* Set timeout value for query operation */
            setPolicydClientTimeout(instance, POLICYD_TIMEOUT);
        }
#endif /* STM */
    }

    event = &(otn->event_data);
    event->userinfo = (struct PolicyUserInfo *)SnortAlloc(sizeof(struct PolicyUserInfo));
    if(event->userinfo == NULL)
    {
        FatalError("%s(%d): Can't Alloc memory for policy user info\n", file_name,
                   file_line);
    }
    
}

void ParseQuery(char *data, OptTreeNode *otn)
{
    QueryPolicyData *ds_ptr;
    char *token, *str, *p;

    ds_ptr = (QueryPolicyData *)otn->ds_list[PLUGIN_QUERYPOLICY];

    while(isspace((int)*data)) data++;
    str = strdup(data);

    if(str == NULL)
    {
        FatalError("ParseQuery: Can't strdup data\n");
    }


    p=str;
    token = strtok(p, ",");

    if(strlen(token) >= 20)
    {
        FatalError("ParseQuery: The protocol name is too long(>=20 bytes");
    }

    strcpy(ds_ptr->policy_protocol, token);

    token = strtok(NULL, ",");
    while(isspace((int)*token)) token++;

    ds_ptr->default_mode = atoi(token);
    if(ds_ptr->default_mode != 0 && ds_ptr->default_mode != 1)
    {
        FatalError("ParseQuery: The default mode is not 0/1"); 
    }

    AddOptFuncToList(QueryPolicy, otn);
    
    free(str);
}

int QueryPolicy(Packet *p, struct _OptTreeNode *otn, OptFpList *fp_list)
{
    QueryPolicyData *ds_ptr;
    POLICYD_ACTION action;
    POLICYD_RET_CODE ret;
    Event *event;

    ds_ptr = (QueryPolicyData *)otn->ds_list[PLUGIN_QUERYPOLICY];

    if(instance == NULL)
    {
        ret = createPolicyClientInstance(&instance);
        if(ret != POLICYD_RET_OK)
        {
            action = POLICYD_ACTION_NO_MATCH;
            instance = NULL;
            ErrorMessage("%s(%d): Create policyd client instance failed, %d\n", 
                          __FILE__, __LINE__, ret);
        }
        else
        {
            action = POLICYD_ACTION_ALLOW;
#ifdef STM
            /* Set timeout value for query operation */
            setPolicydClientTimeout(instance, POLICYD_TIMEOUT);
#endif /* STM */
        }
    }
    else
    {
        action = POLICYD_ACTION_ALLOW;
    }

     
    if(action == POLICYD_ACTION_ALLOW)
    {
        event = &(otn -> event_data);
        if(event->userinfo != NULL)
        {
            /* Init the userinfo and clear the old data  */
            memset(event->userinfo, '\0', sizeof(struct PolicyUserInfo));
        }
#ifdef UTM
        if(p->direction == TO_WAN)
        {
            /* Packet from LAN, use source IP to query. */
            ret = queryAppPolicy(instance, ds_ptr->policy_protocol, 
                                 inet_ntoa(p->iph->ip_src), &action, 
                                 event->userinfo);

        }
        else if(p->direction == FROM_WAN)
        {
            /* Packet from WAN, use dest IP to query. */
            ret = queryAppPolicy(instance, ds_ptr->policy_protocol, 
                                 inet_ntoa(p->iph->ip_dst), &action, 
                                 event->userinfo);
        }
        else
        {
            /*
            ** Can't get the direction, maybe the VPN to WAN traffic,
            ** do not query.
            ** So we assign the POLICYD_RET_OK to ret and treat the actionas
            ** POLICYD_ACTION_NO_MATCH.
            */
            ret = POLICYD_RET_OK;
            action = POLICYD_ACTION_NO_MATCH;
        }
#else
        /* 
        ** In STM, we transfer source IP and dest IP to policyd, 
        ** and let policyd to detemine which IP is LAN side.
        */
	
       char *source_ip, *dest_ip;
       source_ip = strdup(inet_ntoa(p->iph->ip_src));
       dest_ip = strdup(inet_ntoa(p->iph->ip_dst));
       ret = queryAppPolicy(instance, ds_ptr->policy_protocol,
                             source_ip, dest_ip, 
                             &action, event->userinfo);
       free(source_ip);
       free(dest_ip);
#endif /* UTM */
        if(ret != POLICYD_RET_OK)
        {
            action = POLICYD_ACTION_NO_MATCH;
            ErrorMessage("%s(%d): Query policyd failed, %d.\n", 
                          __FILE__, __LINE__, ret);
        }
    }

#ifdef DEBUG
    printf("Query match %s, %s, %d, %d\n", ds_ptr->policy_protocol, inet_ntoa(p->iph->ip_src), ds_ptr->default_mode, action);   
#endif 

    if(action == POLICYD_ACTION_BLOCK 
       || (action == POLICYD_ACTION_UNAUTHORIZED && ds_ptr->default_mode == 1)
       || (action == POLICYD_ACTION_NO_MATCH && ds_ptr->default_mode == 1))
    {
        return fp_list->next->OptTestFunc(p, otn, fp_list->next);
    }
    
    return 0; 

}
#endif
