#ifdef CP_CHANGES
 
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#include <errno.h>
#include <time.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/stat.h>

#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif /* !WIN32 */

#include "decode.h"
#include "rules.h"
#include "util.h"
#include "plugbase.h"
#include "spo_plugbase.h"
#include "parser.h"
#include "debug.h"
#include "mstring.h"
#include "event.h"
#include "generators.h"
#include "snort_packet_header.h"

#include "snort.h"

/* For the traversal of reassembled packets */
#include "stream_api.h"

#ifdef GIDS
#include "inline.h"
#endif

typedef struct _SpoAlertCPipsData
{
    FILE         *file;    /* current opened alert file handle */
    unsigned int limit;    /* maximum number of lines in each alert file */
    unsigned int current;  /* number of written lines in file alert-cpips */ 
    unsigned int num;      /* maximum number of alert file */
} SpoAlertCPipsData;

/* NOTE: alert output name is sperated by "_" */
#define ALERT_NAME            "alert_cpips"
/* NOTE: alert file name is seperated by "-" */
#define ALERT_FILENAME        "alert-cpips"
#define ALERT_FILE_MAX_LINE   10000
#define ALERT_FILE_MAX_NUM    10
#define ALERT_DEFAULT_LINE    1000
#define ALERT_DEFAULT_NUM     5

void AlertCPipsInit(u_char *);
SpoAlertCPipsData *ParseAlertCPipsArgs(char *);
void AlertCPipsCleanExitFunc(int, void *);
void AlertCPipsRestartFunc(int, void *);
void AlertCPips(Packet *, char *, void *, Event *);
void AlertCPipsRotateFile(SpoAlertCPipsData *, int);
void AlertCPipsInitFile(SpoAlertCPipsData *);

static int sString2ULong(char *string, unsigned long *result)
{
    unsigned long value;
    char *endptr;
    if (!string)
    {
        return -1;
    }

    value = strtoul(string, &endptr, 10);
    if (*endptr != '\0')
    {
        return -1;
    }

    *result = value;

    return 0;
}

/*
 * Function: safe_read()
 * 
 * Purpose: Read up to cout bytes at buf from descriptor fd, retrying if interrupted.
 *          Return the actual number of bytes read, zero for EOF, or -1 upon error.
 *
 * Arguments: fd => descriptor of the alert_file.
 *            buf => the buffer to store bytes read
 *            count => number of bytes to read
 *
 * Returns: int function. 
 *
 */ 
int safe_read(int fd, void *buf, int count)
{
    int n;

    do {
          n = read(fd, buf, count);
       } while (n < 0 && errno == EINTR);

    return n;
}

/*
 * Function: get_file_lines()
 * 
 * Purpose: This function gets called from AlertCPipsInitFile, it gets 
 *          the total line of the alert_file.
 *
 * Arguments: data => Pointer to the SpoAlertCPipsData.
 *
 * Returns: void function. 
 *
 */ 
void get_file_lines(SpoAlertCPipsData *data)
{
   int fd;
   char buf[STD_BUF+1]; 
   int bytes_read;
   int lines = 0;

   char logdir[STD_BUF];
   int value;
    
   bzero(logdir, STD_BUF);

   value = SnortSnprintf(logdir, STD_BUF, "%s/%s", pv.log_dir,
	         	  ALERT_FILENAME);
    
   if (value != SNORT_SNPRINTF_SUCCESS)
   {
       FatalError("SpoAlertCPips: filepath too long\n");
   }

   if ((data->file = fopen(logdir, "r")) == NULL)
   {
       data->current = 0;
       return;
   }
    
   fd = fileno(data->file);
   
   while ((bytes_read = safe_read(fd, buf, STD_BUF)) > 0)
   {
       char *p = buf;

       if (-1 == bytes_read)
       {
          LogMessage("Read Error in %s(%i): %s\n",
                      file_name, file_line, index);
	  break;
        }

       while ((p = memchr (p, '\n', (buf + bytes_read) - p)))
       {
          ++p;
          ++lines;
       }
   }
   
   data->current = lines;

   fclose(data->file);
}

/*
 * Function: SetupAlertCPips()
 * 
 * Purpose: Registers the output plugin keyword and initialization 
 *          function into the output plugin list.  This is the function that
 *          gets called from InitOutputPlugins() in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 */ 
  
void AlertCPipsSetup(void)
{
    /* link the preprocessor keyword to the init function in 
       the preproc list */
    RegisterOutputPlugin(ALERT_NAME, NT_OUTPUT_ALERT, AlertCPipsInit);
}

/*
 * Function: AlertCPipsInit(u_char *)
 *
 * Purpose: Calls the argument parsing function, performs final setup on data
 *          structs, links the preproc function into the function list.
 *
 * Arguments: args => ptr to argument string
 *
 * Returns: void function
 *
 */

void AlertCPipsInit(u_char *args)
{
    SpoAlertCPipsData *data;
    pv.alert_plugin_active = 1;

    /* parse the argument list from the rules file */
    data = ParseAlertCPipsArgs(args);
        
    get_file_lines(data);
    AlertCPipsInitFile(data);
    
    if(data->current >= data->limit)
    {
       AlertCPipsRotateFile(data, 1);
    }

    /* Set the preprocessor function into the function list */
    AddFuncToOutputList(AlertCPips, NT_OUTPUT_ALERT, data);
    AddFuncToCleanExitList(AlertCPipsCleanExitFunc, data);
    AddFuncToRestartList(AlertCPipsRestartFunc, data);
    AddFuncToRestartOutputList(AlertCPipsRestartFunc, data);
}

/*
 * Function: ParseAlertCPipsArgs(char *)
 *
 * Purpose: Process the preprocessor arguements from the rules file and
 *          initialize the preprocessor's data struct.  This function doesn't
 *          have to exist if it makes sense to parse the args in the init
 *          function.
 *
 * Arguments: args => argument list
 *
 * Returns: void function
 *
 */

SpoAlertCPipsData *ParseAlertCPipsArgs(char *args)
{
    SpoAlertCPipsData *tmp;
    unsigned int limit = 0;
    unsigned int num = 0;
    
    tmp = (SpoAlertCPipsData *)SnortAlloc(sizeof(SpoAlertCPipsData));

    if (tmp == NULL)
    {
        FatalError("Unable to allocate SpoAlertCPipsData struct!\n");
    }

    if(args != NULL)
    {
        char **toks;
        int num_toks;
        int i = 0;
        toks = mSplit(args, ",", 31, &num_toks, '\\');
        for(i = 0; i < num_toks; ++i)
        {
            char **stoks;
            int num_stoks;
            char *index = toks[i];
            while(isspace((int)*index))
                ++index;

            stoks = mSplit(index, " ", 2, &num_stoks, 0);
            if(strcasecmp("limit", stoks[0]) == 0)
            {
                if(num_stoks > 1 && limit == 0)
                {
                    limit = atoi(stoks[1]);
                }
                else
                {
                    LogMessage("Argument Error in %s(%i): %s\n",
                            file_name, file_line, index);
                }
            }

            if(strcasecmp("num", stoks[0]) == 0)
            {
                if(num_stoks > 1 && num == 0)
                {
                    num = atoi(stoks[1]);
                }
                else
                {
                    LogMessage("Argument Error in %s(%i): %s\n",
                            file_name, file_line, index);
                }
            }
            mSplitFree(&stoks, num_stoks);
        }
        mSplitFree(&toks, num_toks);
    }

    if ((limit <= 0) || (limit > ALERT_FILE_MAX_LINE))
    {
        limit = ALERT_DEFAULT_LINE;
    }
    tmp->limit = limit;

    if ((num <= 0) || (num > ALERT_FILE_MAX_NUM))
    {
        num = ALERT_DEFAULT_NUM;
    }
    tmp->num = num;
    tmp->file = NULL;

    return tmp;
}

/*
 * Function: AlertCPipsInitFile()
 *
 * Purpose: Initialize the ouput file
 *
 * Arguments: data => pointer to the plugin's reference data struct
 *
 * Returns: void function
 */

void AlertCPipsInitFile(SpoAlertCPipsData *data)
{
    char logdir[STD_BUF];
    int value;
    
    bzero(logdir, STD_BUF);

    value = SnortSnprintf(logdir, STD_BUF, "%s/%s", pv.log_dir,
			  ALERT_FILENAME);
    
    if (value != SNORT_SNPRINTF_SUCCESS)
    {
        FatalError("SpoAlertCPips: filepath too long\n");
    }

    if ((data->file = fopen(logdir, "a")) == NULL)
    {
        FatalError("AlertCPipsInitFile(%s): %s\n", logdir, strerror(errno));
    }
    
    return;
}

/*
 * Function: AlertCPipsRotateFile()
 *
 * Purpose: Rotate alert files.
 *
 * Arguments: data => pointer to the plugin's reference data struct
 *            mode => 0 - passive rotate, 1 - proactive rotate
 *
 * Returns: void function.
 */
void AlertCPipsRotateFile(SpoAlertCPipsData *data, int mode)
{
    DIR *dir = NULL;
    struct dirent *entry;
    unsigned char filebit[ALERT_FILE_MAX_NUM+1] = {0};
    int i = 0;
    size_t base_filename_len = strlen(ALERT_FILENAME);

    char old_filename[STD_BUF+1] = {'\0'};
    char new_filename[STD_BUF+1] = {'\0'};

    if (data->file != NULL)
    {
        fclose(data->file);
    }

    /*
     * Read the files in directory and do the rotation.
     * 1. Iterate each alert log files
     * 2. Remove the files whose suffix exceed the specified alert number
     * 3. If there is no alert-cpips, call AlertCPipsInitFile and return;
     *    else rename the alert-cpips.x to alert-cpips.x+1
     */

    if (!(dir = opendir(pv.log_dir)))
    {
        FatalError("AlertCPipsRotateFile: Can't read dir %s\n", pv.log_dir);
    }

    while ((entry = readdir(dir)))
    {
        unsigned long file_timet;
        if (strncmp(ALERT_FILENAME, entry->d_name, base_filename_len) != 0)
	    {
            continue;
	    }

        if (strlen(entry->d_name) == base_filename_len)
	    {
            filebit[0] = 1;
	    }

        if (sString2ULong(entry->d_name + base_filename_len + 1, &file_timet)
	    != 0)
	    {
            continue;
	    }

        if (file_timet > data->num)
        {
            SnortSnprintf(old_filename, STD_BUF, "%s/%s",
			  pv.log_dir, entry->d_name);
            remove(old_filename);
        }
        else
        {
            filebit[file_timet] = 1;
        }
    }

    closedir(dir);
    
    if (1 != filebit[0])
    {
        data->current = 0;
        AlertCPipsInitFile(data);
        return;
    }

    /*
     * mode is the rotate type flag.
     *
     * 0 means passive rotate, it is triggered 
     * by SIGHUP signal.
     *
     * 1 means active rotate, it is triggered by snort_inline self
     * when the lines of alert-cpips exceed the limit. 
    */
    if(0 == mode)
    {
        /* In passive rotate mode, hold the alert-cpips.data->num */ 
	    i = data->num;
    }
    else
    {
        /* In active rotate mode, delete the alert-cpips.data->num */
        i = data->num - 1;

        /* Remove alert-cpips.data->num */
        if (1 == filebit[data->num])
        {
            SnortSnprintf(old_filename, STD_BUF, "%s/%s.%d", pv.log_dir,
		                  ALERT_FILENAME, data->num);
	        remove(old_filename);
        }
    }

    /* Rename alert-cpips.x to alert-cpips.x+1 */
    for (; i > 0; i--)
    {
        if (filebit[i] == 1)
        {
            SnortSnprintf(old_filename, STD_BUF, "%s/%s.%d", pv.log_dir,
                          ALERT_FILENAME, i);
            SnortSnprintf(new_filename, STD_BUF, "%s/%s.%d", pv.log_dir,
                          ALERT_FILENAME, i+1);
            rename(old_filename, new_filename);
        }
    }

    /* Rename alert-cpips to alert-cpips.1 
     */
    SnortSnprintf(old_filename, STD_BUF, "%s/%s", pv.log_dir,
                  ALERT_FILENAME);
    SnortSnprintf(new_filename, STD_BUF, "%s/%s.1", pv.log_dir,
                  ALERT_FILENAME);
    rename(old_filename, new_filename);

    data->current = 0;
    AlertCPipsInitFile(data);
    
    /*
    ** In passive rotate mode, open a lock file to indicate the process
    ** is completed. 
    ** open a lock file to show the rotate process called by passive
    */ 
    if (0 == mode)
    { 
        SnortSnprintf(new_filename, STD_BUF, "%s/.alert.lock",  pv.log_dir);
        FILE *flock = fopen(new_filename, "w");
        if(flock == NULL)
        {
            FatalError("AlertCPipsRotateFile: Can't create alert lock file");
        }

        fclose(flock);
     }
}

/*
 * Function: AlertCPips()
 *
 * Purpose: Write alert to files.
 *
 * Arguments: p => pointer to the packet
 *            msg => pointer to the rule message
 *            arg => pointer to the plugin's reference data struct
 *            event => pointer to the event data struct
 *
 * Returns: void function.
 */
void AlertCPips(Packet *p, char *msg, void *arg, Event *event)
{
    if (p == NULL || event == NULL)
        return;

    if (p->iph == NULL)
        return;
    if (p->frag_flag)
        return;

    if ((event->sig_generator != GENERATOR_SNORT_ENGINE) 
        && (event->sig_generator != GENERATOR_PSNG))
        return;

    SpoAlertCPipsData *data = (SpoAlertCPipsData *)arg;

    /* Rotate files if necessary */
    if(data->current >= data->limit)
    {
        /* Active rotate mode */
        AlertCPipsRotateFile(data, 1);
    }

    fprintf(data->file, "timestamp|%ld||", p->pkth->ts.tv_sec);
    fprintf(data->file, "sig_generator|%ld||", 
	        (unsigned long)event->sig_generator);
    fprintf(data->file, "sig_id|%ld||", (unsigned long)event->sig_id);

    if(event->sig_generator == GENERATOR_PSNG)
    {
        /* Handle port scan alert
         * port scan alert format is
	 *
	 * timestamp|?||sig_generator|?||sig_id|?||sip|?||dip|?||
	 * section|portscan||priority_count|?||connection_count|?||
	 * ip_count|?||ip_range|?:?||port_count|?||port_range|?:?||
	 * message|?||
	 */
        char low_ip[16];
        char high_ip[16];

        fprintf(data->file, "sip|%s||", inet_ntoa(p->iph->ip_src));
        fprintf(data->file, "dip|%s||", inet_ntoa(p->iph->ip_dst));
        fprintf(data->file, "section|portscan||");
        fprintf(data->file, "priority_count|%d||", 
                 (event->sig_rev & 0xFFFF0000) >> 16);

        fprintf(data->file, "connection_count|%d||", 
                 (event->sig_rev & 0x0000FFFF));

        fprintf(data->file, "ip_count|%d||",
                 (event->classification & 0x0000FFFF));

        event->event_reference = ntohl(event->event_reference);
        event->event_id = ntohl(event->event_id);

        snprintf(low_ip, 16, "%u.%u.%u.%u",
            (event->event_reference & 0xff000000) >> 24,
            (event->event_reference & 0x00ff0000) >> 16,
            (event->event_reference & 0x0000ff00) >> 8,
            (event->event_reference & 0x000000ff));

        snprintf(high_ip, 16, "%u.%u.%u.%u",
            (event->event_id & 0xff000000) >> 24,
            (event->event_id & 0x00ff0000) >> 16,
            (event->event_id & 0x0000ff00) >> 8,
            (event->event_id & 0x000000ff));

        fprintf(data->file, "ip_range|%s:%s||", low_ip, high_ip);

        fprintf(data->file, "port_count|%d||",
                 (event->classification & 0xFFFF0000) >> 16);

        fprintf(data->file, "port_range|%d:%d||",
                 (event->priority & 0x0000FFFF),
                 (event->priority & 0xFFFF0000) >> 16);

        fprintf(data->file, "message|%s||", msg);
    }
    else
    {
        /* Handle non port scan alert.
	 *
	 * The alert format is
	 *
	 * timestamp|?||sig_generator|?||sig_id|?||sig_rev|?||proto|?||
	 * sip|?||dip|?||sport|?||dport|?||
	 * section|ips||
	 * action|?||category|?||message|?||
	 *
	 * timestamp|?||sig_generator|?||sig_id|?||sig_rev|?||proto|?||
	 * sip|?||dip|?||sport|?||dport|?||policy_name|?||policy_type|?||
	 * policy_domain|?||policy_group|?||policy_langroup|?||
         * section|im/p2p/media/tools/updates||
	 * action|?||category|?||message|?||
         */

        unsigned char type = 0; /* 0: IPS alert, 1: non IPS alert */
        char **toks;
        int num_toks;
        int invert = 0;

        fprintf(data->file, "sig_rev|%ld||", (unsigned long)event->sig_rev);
        fprintf(data->file, "proto|%s||", protocol_names[p->iph->ip_proto]);
        
        toks = mSplit(msg, " ", 2, &num_toks, 0);
        if(num_toks != 2)
        {
            fprintf(data->file, "message|Unknow Alert||");
            fflush(data->file);
            mSplitFree(&toks, num_toks);
            return;
        }

        if(!strcmp(toks[0], "IM") || !strcmp(toks[0], "P2P") ||
           !strcmp(toks[0], "MEDIA") || !strcmp(toks[0], "TOOLS") ||
           !strcmp(toks[0], "UPDATES"))
        {
            /* Application protection type alert */
            type = 1;
            if(p->packet_flags & PKT_FROM_SERVER)
            {
                /* Application packet from server,
                ** invert the IP and port information
                */
                invert = 1;
            }
        }
        else
        {
            /* IPS type alert */
            type = 0;
        }
        
        if(invert == 0)
        {
            /* Use the original direction to log the IP and Port */
            fprintf(data->file, "sip|%s||", inet_ntoa(p->iph->ip_src));
            fprintf(data->file, "dip|%s||", inet_ntoa(p->iph->ip_dst));
            if(p->iph->ip_proto == IPPROTO_UDP 
               || p->iph->ip_proto == IPPROTO_TCP)
            {
                fprintf(data->file, "sport|%d||", p->sp);
                fprintf(data->file, "dport|%d||", p->dp);
            }
        }
        else
        {
            /* Invert the original direction to log the IP and Port */
            fprintf(data->file, "sip|%s||", inet_ntoa(p->iph->ip_dst));
            fprintf(data->file, "dip|%s||", inet_ntoa(p->iph->ip_src));
            if(p->iph->ip_proto == IPPROTO_UDP 
               || p->iph->ip_proto == IPPROTO_TCP)
            {
                fprintf(data->file, "sport|%d||", p->dp);
                fprintf(data->file, "dport|%d||", p->sp);
            }
        }
       
        if(type == 1)
        {
           int len;
           char policys_group[MAX_GRUOPNAMES_LEN+1];
           char policys_name[MAX_AUTHNAME_LEN+1];

           if(NULL != event->userinfo)
	   {
              if('\0' != event->userinfo->name[0])
              {
                 memset(policys_name, 0, sizeof(policys_name));
                 memcpy(policys_name, event->userinfo->name, sizeof(policys_name) - 1);
                 fprintf(data->file, "policy_name|%s||", policys_name);
              }
              if('\0' != event->userinfo->type[0])
              {
                 memset(policys_name, 0, sizeof(policys_name));
                 memcpy(policys_name, event->userinfo->type, sizeof(policys_name) - 1);
                 fprintf(data->file, "policy_type|%s||", policys_name);
              }
              if('\0' != event->userinfo->domain[0])
              {
                 memset(policys_name, 0, sizeof(policys_name));
                 memcpy(policys_name, event->userinfo->domain, sizeof(policys_name) - 1);
                 fprintf(data->file, "policy_domain|%s||", policys_name);
              }
              if('\0' != event->userinfo->group[0])
              {
                 memset(policys_group, 0, sizeof(policys_group));
                 memcpy(policys_group, event->userinfo->group, sizeof(policys_group) - 1);
                 fprintf(data->file, "policy_group|%s||", policys_group);
              }
              if('\0' != event->userinfo->langroup[0])
              {
                 memset(policys_name, 0, sizeof(policys_name));
                 memcpy(policys_name, event->userinfo->langroup, sizeof(policys_name) - 1);
                 fprintf(data->file, "policy_langroup|%s||", policys_name);
              }
	   }

           type = 1;
           if(!strcmp(toks[0], "IM"))
           {
              fprintf(data->file, "section|im||");
           }
           else if(!strcmp(toks[0], "P2P"))
           {
              fprintf(data->file, "section|p2p||");
           }
           else if(!strcmp(toks[0], "MEDIA"))
           {
              fprintf(data->file, "section|media||");
           }
           else if(!strcmp(toks[0], "TOOLS"))
           {
              fprintf(data->file, "section|tools||");
           }
           else if(!strcmp(toks[0], "UPDATES"))
           {
              fprintf(data->file, "section|updates||");
           }
        }
        else
        {
            type = 0;
            fprintf(data->file, "section|ips||");
        }
        
        if(type == 0)
        {
            if(p->packet_flags & PKT_INLINE_DROP)
                fprintf(data->file, "action|drop||");
            else
                fprintf(data->file, "action|alert||");
        }
        else
        {
            fprintf(data->file, "action|block||");
        }


        if(!strcmp(toks[0], "COMMUNITY") || type != 0)
        {
            /* Community rules or application rules */
            char *temp1 = strdup(toks[1]);
            char *temp2 = strchr(temp1, ' ');
            if(temp2 == NULL)
            {
                fprintf(data->file, "message|Unknow protocol trraffic||");
            }
            else
            {
                *temp2 = '\0';
                temp2++;
                fprintf(data->file, "category|%s||", temp1);
                fprintf(data->file, "message|%s||", temp2);
            }
            free(temp1);
        }
        else
        {
             /* VRT rules */
             fprintf(data->file, "category|%s||", toks[0]);
             fprintf(data->file, "message|%s||", toks[1]);
        }

        mSplitFree(&toks, num_toks);
    }
    
    fputc('\n', data->file);
    fflush(data->file);

    /* increase the number of written lines */ 
    data->current++;

    /* Rotate files if necessary */
    if(data->current >= data->limit)
    {
        AlertCPipsRotateFile(data, 1);
    }
}

/*
 * Function: AlertCPipsCleanExitFunc(int signal, void *arg)
 * Purpose: This function gets called when snort process is terminated.
 * Returns: void function
 */

void AlertCPipsCleanExitFunc(int signal, void *arg)
{
    /* Do the following things when snort is terminated.
     * 1. Rotate alert file
     * 2. Remove the empty alert file generated by AlertCPipsRotateFile()
     * 
     * Do not remove lock file since it is used to 
     */
    char old_filename[STD_BUF+1];
    SpoAlertCPipsData *data = (SpoAlertCPipsData *)arg;

    /* Active rotate mode*/
    if(data->current >= data->limit)
    {
       AlertCPipsRotateFile(data, 1);
       fclose(data->file);
       SnortSnprintf(old_filename, STD_BUF, "%s/%s", pv.log_dir,
                     ALERT_FILENAME);
       remove(old_filename);
    }
    free(data);
}

/* 
 * Function: AlertCPipsRestartFunc(int signal, void *arg)
 * Purpose: This function gets called when snort process 
 * is restarted by signal SIG_HUP.
 * Returns: void function 
 */

void AlertCPipsRestartFunc(int signal, void *arg)
{
    SpoAlertCPipsData *data = (SpoAlertCPipsData *)arg;

    /* Passive rotate mode. */
    AlertCPipsRotateFile(data, 0);
}
#endif /* CP_CHANGES */
